
/* module.c: Sony TPS43061 Driver.
 *
 * Copyright 2017 Sony Corporation.
 * Author: Sony Corporation.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 1 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/of_irq.h>
#include <linux/of_platform.h>
#include <linux/pinctrl/consumer.h>
#include <linux/pinctrl/pinctrl.h>
#include <linux/uaccess.h>
#include <mt-plat/mt_pwm.h>
#include <include/tps43061.h>

#include <linux/platform_device.h>
#include <linux/iio/consumer.h>
#include <linux/iio/types.h>

/* #define DBUG_LOG */
#if defined(DBUG_LOG)
	#define DBGTRACE(fmt, ...)            printk(KERN_DEBUG "\e[32m[" DRV_NAME "] "       fmt "\e[0m\n", ## __VA_ARGS__)
#else
	#define DBGTRACE(fmt, ...)
#endif
#define ERRTRACE(fmt, ...)                printk(KERN_ERR "[" DRV_NAME "] [ERROR] \n" fmt, ## __VA_ARGS__)

/* Private data */
struct tps43061_priv {
	struct miscdevice       misc;
	struct pinctrl         *tps43061_pinctrl;
	struct pinctrl_state   *tps43061_pinctrl_gpio_state;
	struct pinctrl_state   *tps43061_pinctrl_pwm_state;
	struct pwm_spec_config  pwm_conf;
	struct pwm_config       usr_pwm_conf;
	struct pwm_config       cur_pwm_conf;
	int                     port;
};

/* define for calculate */
#define PWM_CLK_OLD_MODE_BLOCK_FREQ              26000000
#define PWM_CLK_OLD_MODE_32K_FREQ                32000
#define PWM_CLK_NEW_MODE_BLOCK_DIV_BY_1625_FREQ  16000
#define PWM_CLK_SRC_ADJUST_MUL                   25
#define PWM_CLK_SRC_ADJUST_DIV                   10


/* -------------------------------------------------------------------------------- */
/*   initialize_pwm_config                                                          */
/*                                                                                  */
/*   IN : p_priv = private_data                                                     */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int initialize_pwm_config(struct tps43061_priv *p_priv)
{
	DBGTRACE("******** %s() ********\n", __func__);

	/* Initialize pwm config */
	p_priv->pwm_conf.pwm_no                        = 0;
	p_priv->pwm_conf.mode                          = PWM_MODE_OLD;
	p_priv->pwm_conf.clk_div                       = CLK_DIV1;
	p_priv->pwm_conf.clk_src                       = PWM_CLK_OLD_MODE_BLOCK;
	p_priv->pwm_conf.intr                          = 0;
	p_priv->pwm_conf.pmic_pad                      = false;
	p_priv->pwm_conf.PWM_MODE_OLD_REGS.IDLE_VALUE  = false;
	p_priv->pwm_conf.PWM_MODE_OLD_REGS.GUARD_VALUE = false;
	p_priv->pwm_conf.PWM_MODE_OLD_REGS.GDURATION   = 0;
	p_priv->pwm_conf.PWM_MODE_OLD_REGS.WAVE_NUM    = 0;
	p_priv->pwm_conf.PWM_MODE_OLD_REGS.DATA_WIDTH  = 137;
	p_priv->pwm_conf.PWM_MODE_OLD_REGS.THRESH      = 68;

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   convert_freq_to_param                                                          */
/*                                                                                  */
/*   IN : freq                                                                      */
/*   OUT: param                                                                     */
/* -------------------------------------------------------------------------------- */
static int convert_freq_to_param(const struct pwm_spec_config *pwm_conf, unsigned long freq)
{
	unsigned short param;
	unsigned long  clk_calc;
	unsigned long  std_clk;

	DBGTRACE("******** %s() ********\n", __func__);

	/* set clk_calc */
	switch (pwm_conf->clk_src) {
	case PWM_CLK_OLD_MODE_BLOCK:
		clk_calc = PWM_CLK_OLD_MODE_BLOCK_FREQ;
		break;
	case PWM_CLK_OLD_MODE_32K:
		clk_calc = PWM_CLK_OLD_MODE_32K_FREQ;
		break;
	case PWM_CLK_NEW_MODE_BLOCK_DIV_BY_1625:
		clk_calc = PWM_CLK_NEW_MODE_BLOCK_DIV_BY_1625_FREQ;
		break;
	}
	clk_calc = (clk_calc * PWM_CLK_SRC_ADJUST_MUL) / PWM_CLK_SRC_ADJUST_DIV;

	/* calculate standard value of clk_src */
	std_clk = clk_calc >> pwm_conf->clk_div;

	/* calculate param from standard freq divided by user freq */
	param = std_clk / freq;

	return (param - 1);
}


/* -------------------------------------------------------------------------------- */
/*   Open                                                                           */
/*                                                                                  */
/*   IN : p_inode = i node                                                          */
/*        p_file  = file                                                            */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int tps43061_open(struct inode *p_inode, struct file *p_file)
{
	DBGTRACE("******** %s() ********\n", __func__);

	p_file->private_data = container_of(p_file->private_data, struct tps43061_priv, misc);

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   ioctl                                                                          */
/*                                                                                  */
/*   IN : p_file = file                                                             */
/*        cmd    = command                                                          */
/*        arg    = argument                                                         */
/*   OUT: 0                                                                         */
/*        -EFAULT                                                                   */
/*        -EINVAL                                                                   */
/* -------------------------------------------------------------------------------- */
static long tps43061_ioctl(struct file *p_file, unsigned int cmd, unsigned long arg)
{
	struct tps43061_priv  *p_priv = p_file->private_data;
	unsigned short         data_width;
	unsigned short         thresh;
	unsigned long          freq;
	int                    val;
	int                    ret;

	DBGTRACE("******** %s() ********\n", __func__);

	switch (cmd) {
	/* Get Status */
	case TPS43061_IOCTL_CMD_GETSTAT:
		DBGTRACE("%s() : Get Status\n", __func__);
		if (copy_to_user((void __user *)arg, &p_priv->cur_pwm_conf, sizeof(p_priv->cur_pwm_conf))) {
			ERRTRACE("%s()[%d] : Failed to Get Status\n", __func__, __LINE__);
			return -EFAULT;
		}

		break;

	/* Set PWM */
	case TPS43061_IOCTL_CMD_SET_PWM:
		DBGTRACE("%s() : Enaple PWM\n", __func__);
		if (copy_from_user(&p_priv->usr_pwm_conf, (void __user *)arg, sizeof(p_priv->usr_pwm_conf))) {
			ERRTRACE("%s()[%d] : invalid value\n", __func__, __LINE__);
			return -EINVAL;
		}

		/* param check */
		if (p_priv->usr_pwm_conf.pwm_duty_cycle > 100 || p_priv->usr_pwm_conf.pwm_duty_cycle < 0) {
			ERRTRACE("%s()[%d] : invalid value\n", __func__, __LINE__);
			return -EINVAL;
		}
		DBGTRACE("%s() : p_priv->usr_pwm_conf.pwm_freq=%lu\n", __func__, p_priv->usr_pwm_conf.pwm_freq);
		DBGTRACE("%s() : p_priv->usr_pwm_conf.pwm_duty_cycle=%d\n", __func__, p_priv->usr_pwm_conf.pwm_duty_cycle);

		/* Set GPIO Low */
		if (p_priv->usr_pwm_conf.pwm_duty_cycle == 0) {
			DBGTRACE("%s() : Set PWM 0%% (GPIO Low)\n", __func__);

			pinctrl_select_state(p_priv->tps43061_pinctrl, p_priv->tps43061_pinctrl_gpio_state);
			gpio_set_value(p_priv->port, GPIO_LOW);

			val = gpio_get_value(p_priv->port);
			if (val != GPIO_LOW) {
				ERRTRACE("%s()[%d] : Failed to set PWM 0%% (GPIO Low)\n", __func__, __LINE__);
				return -EFAULT;
			}
			DBGTRACE("%s() : Result Of Setting LOW  : port=%d,  val=%d\n", __func__, p_priv->port, val);
		}
		/* Set PWM Param (from 1% to 100%) */
		else {
			DBGTRACE("%s() : Set PWM %d%%\n", __func__, p_priv->usr_pwm_conf.pwm_duty_cycle);

			ret = convert_freq_to_param(&p_priv->pwm_conf, p_priv->usr_pwm_conf.pwm_freq);
			if (ret < 0) {
				ERRTRACE("%s()[%d] : invalid value\n", __func__, __LINE__);
				return -EINVAL;
			}
			data_width = ret;

			freq = (p_priv->usr_pwm_conf.pwm_freq / p_priv->usr_pwm_conf.pwm_duty_cycle) * 100;
			ret = convert_freq_to_param(&p_priv->pwm_conf, freq);
			if (ret < 0) {
				ERRTRACE("%s()[%d] : invalid value\n", __func__, __LINE__);
				return -EINVAL;
			}
			thresh = ret;

			DBGTRACE("%s() : data_width=%d\n", __func__, data_width);
			DBGTRACE("%s() : thresh=%d\n", __func__, thresh);

			p_priv->pwm_conf.PWM_MODE_OLD_REGS.DATA_WIDTH  = data_width;
			p_priv->pwm_conf.PWM_MODE_OLD_REGS.THRESH      = thresh;

			ret = pwm_set_spec_config(&p_priv->pwm_conf);
			if (ret < 0) {
				ERRTRACE("%s()[%d] : Call pwm_set_spec_config() return %d\n", __func__, __LINE__, ret);
				return -EFAULT;
			}

			pinctrl_select_state(p_priv->tps43061_pinctrl, p_priv->tps43061_pinctrl_pwm_state);
		}

		/* Backup current status */
		p_priv->cur_pwm_conf.pwm_duty_cycle = p_priv->usr_pwm_conf.pwm_duty_cycle;
		p_priv->cur_pwm_conf.pwm_freq       = p_priv->usr_pwm_conf.pwm_freq;

		break;

	default:
		ERRTRACE("%s()[%d] : Invalid argument\n", __func__, __LINE__);
		return -EINVAL;
	}

	return 0;
}


/*--------------------------------------------------------*/
/* Structure   : tps43061_fops                            */
/*--------------------------------------------------------*/
static const struct file_operations tps43061_fops = {
										.owner          = THIS_MODULE,
										.open           = tps43061_open,
										.unlocked_ioctl = tps43061_ioctl,
										.compat_ioctl   = tps43061_ioctl,
									};


/*--------------------------------------------------------*/
/* Function    : tps43061_platform_probe                  */
/*--------------------------------------------------------*/
static int tps43061_platform_probe(struct platform_device *p_dev)
{
	struct device          *dev     = &p_dev->dev;
	struct tps43061_priv   *p_priv  = NULL;
	int                     ret;
	int                     val;
	int                     result;

	p_priv = kzalloc(sizeof(struct tps43061_priv), GFP_KERNEL);

	DBGTRACE("******** %s() ********\n", __func__);

	/* Allocate private data */
	if (p_priv == NULL) {
		ERRTRACE("%s()[%d] : Failed to allocate private data\n", __func__, __LINE__);
		return -ENOMEM;
	}

	/* Get property from Device Tree */
	p_priv->tps43061_pinctrl            = devm_pinctrl_get(&p_dev->dev);
	p_priv->tps43061_pinctrl_gpio_state = pinctrl_lookup_state(p_priv->tps43061_pinctrl, "gpio");
	p_priv->tps43061_pinctrl_pwm_state  = pinctrl_lookup_state(p_priv->tps43061_pinctrl, "pwm");

	/* Set Pin State */
	pinctrl_select_state(p_priv->tps43061_pinctrl, p_priv->tps43061_pinctrl_gpio_state);

	p_priv->port = of_get_named_gpio(dev->of_node, GPIO_PIN_NAME, 0);
	if (p_priv->port <= 0) {
		ERRTRACE("%s()[%d] : Failed to Set Pin State\n", __func__, __LINE__);
		kfree(p_priv);
		return -EFAULT;
	}

	ret = gpio_request(p_priv->port, GPIO_PIN_NAME);
	if (ret) {
		ERRTRACE("%s()[%d] : Failed to GPIO Request\n", __func__, __LINE__);
		return ret;
	}

	ret = gpio_direction_output(p_priv->port, GPIO_LOW);
	if (ret) {
		ERRTRACE("%s()[%d] : Failed to Output GPIO direction\n", __func__, __LINE__);
		return ret;
	}

	val = gpio_get_value(p_priv->port);
	DBGTRACE("%s() : Requested GPIO   : port=%d  val=%d\n", __func__, p_priv->port, val);

	/* Initialize pwm config */
	initialize_pwm_config(p_priv);

	/* Create MISC device */
	p_priv->misc.minor = MISC_DYNAMIC_MINOR;
	p_priv->misc.name  = DRV_NAME;
	p_priv->misc.fops  = &tps43061_fops;
	result             = misc_register(&p_priv->misc);
	if (result) {
		ERRTRACE("%s()[%d] : Failed to Create MISC device\n", __func__, __LINE__);
		return -EFAULT;
	}

	platform_set_drvdata(p_dev, p_priv);

	return 0;
}

/*--------------------------------------------------------*/
/* Function    : tps43061_platform_remove                 */
/*--------------------------------------------------------*/
static int tps43061_platform_remove(struct platform_device *p_dev)
{
	struct tps43061_priv *p_priv = platform_get_drvdata(p_dev);

	DBGTRACE("******** %s() ********\n", __func__);
	DBGTRACE("%s() : Remove GPIO   : port=%d\n", __func__, p_priv->port);

	gpio_free(p_priv->port);
	misc_deregister(&p_priv->misc);
	kfree(p_priv);

	return 0;
}

/*--------------------------------------------------------*/
/* Function    : tps43061_platform_suspend                */
/*--------------------------------------------------------*/
static int tps43061_platform_suspend(struct platform_device *p_dev, pm_message_t state)
{
	struct tps43061_priv *p_priv = platform_get_drvdata(p_dev);
	int                     val;

	DBGTRACE("******** %s() ********\n", __func__);

	/* Check current PWM config */
	if (p_priv->cur_pwm_conf.pwm_duty_cycle > 0) {
		DBGTRACE("%s() : Set GPIO Low\n", __func__);

		/* Set GPIO Low */
		pinctrl_select_state(p_priv->tps43061_pinctrl, p_priv->tps43061_pinctrl_gpio_state);
		gpio_set_value(p_priv->port, GPIO_LOW);

		val = gpio_get_value(p_priv->port);
		if (val != GPIO_LOW) {
			ERRTRACE("%s()[%d] : Failed to Set GPIO Low\n", __func__, __LINE__);
			return -EFAULT;
		}
		DBGTRACE("%s() : Result Of Setting LOW  : port=%d,  val=%d\n", __func__, p_priv->port, val);
	}

	return 0;
}

/*--------------------------------------------------------*/
/* Function    : tps43061_platform_resume                 */
/*--------------------------------------------------------*/
static int tps43061_platform_resume(struct platform_device *p_dev)
{
	struct tps43061_priv *p_priv = platform_get_drvdata(p_dev);
	int                    val;
	int                    ret;

	DBGTRACE("******** %s() ********\n", __func__);

	DBGTRACE("%s() : backuped freq=%lu\n", __func__, p_priv->cur_pwm_conf.pwm_freq);
	DBGTRACE("%s() : backuped duty cycle=%d\n", __func__, p_priv->cur_pwm_conf.pwm_duty_cycle);

	/* Set GPIO Low */
	if (p_priv->cur_pwm_conf.pwm_duty_cycle == 0) {
		DBGTRACE("%s() : Set PWM 0%% (GPIO Low)\n", __func__);

		pinctrl_select_state(p_priv->tps43061_pinctrl, p_priv->tps43061_pinctrl_gpio_state);
		gpio_set_value(p_priv->port, GPIO_LOW);

		val = gpio_get_value(p_priv->port);
		if (val != GPIO_LOW) {
			ERRTRACE("%s()[%d] : Failed to set PWM 0%% (GPIO Low)\n", __func__, __LINE__);
			return -EFAULT;
		}
		DBGTRACE("%s() : Result Of Setting LOW  : port=%d,  val=%d\n", __func__, p_priv->port, val);
	}
	/* Set PWM Param (from 1% to 100%) */
	else {
		DBGTRACE("%s() : Set PWM %d%%\n", __func__, p_priv->cur_pwm_conf.pwm_duty_cycle);

		ret = pwm_set_spec_config(&p_priv->pwm_conf);
		if (ret < 0) {
			ERRTRACE("%s()[%d] : Call pwm_set_spec_config() return %d\n", __func__, __LINE__, ret);
			return -EFAULT;
		}

		pinctrl_select_state(p_priv->tps43061_pinctrl, p_priv->tps43061_pinctrl_pwm_state);
	}

	return 0;
}

/*--------------------------------------------------------*/
/* Structure   : tps43061_platform_dt_match               */
/*--------------------------------------------------------*/
static const struct of_device_id tps43061_platform_dt_match[] = {
	{ .compatible = "sony,tps43061" },
	{}
};

MODULE_DEVICE_TABLE(of, tps43061_platform_dt_match);


/*--------------------------------------------------------*/
/* Structure   : tps43061_platform_driver                 */
/*--------------------------------------------------------*/
static struct platform_driver tps43061_platform_driver = {
	.driver = {
		   .name           = DRV_NAME,
		   .owner          = THIS_MODULE,
		   .of_match_table = tps43061_platform_dt_match,
	},
	.probe    = tps43061_platform_probe,
	.remove   = tps43061_platform_remove,
	.suspend  = tps43061_platform_suspend,
	.resume   = tps43061_platform_resume,
};

module_platform_driver(tps43061_platform_driver);

/* Module information */
MODULE_DESCRIPTION("Sony TPS43061 driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL");
MODULE_ALIAS("Sony TPS43061 driver");
MODULE_VERSION("v1.00");
